@@ -6,6 +6,8 @@ from itertools import chain, islice
 from numbers import Integral
 import sys
 import re
+import warnings
+import operator
 
 import six
 
@@ -18,13 +20,9 @@ BRANCH_FACTOR = 32
 BIT_MASK = BRANCH_FACTOR - 1
 SHIFT = _bitcount(BIT_MASK)
 
-def _comparator(f):
-    @wraps(f)
-    def wrapper(*args, **kwds):
-        if isinstance(args[0], PVector) and isinstance(args[1], PVector):
-            return f(*args, **kwds)
-        return NotImplemented
-    return wrapper
+
+def compare_pvector(v, other, operator):
+    return operator(v.tolist(), other.tolist() if isinstance(other, PVector) else other)
 
 
 class _PVectorImpl(object):
@@ -77,29 +75,23 @@ class _PVectorImpl(object):
         # by far of those tried since it uses the speed of the built in python list directly.
         return iter(self.tolist())
 
-    @_comparator
     def __ne__(self, other):
-        return self.tolist() != other.tolist()
+        return compare_pvector(self, other, operator.ne)
 
-    @_comparator
     def __eq__(self, other):
-        return self is other or self.tolist() == other.tolist()
+        return self is other or compare_pvector(self, other, operator.eq)
 
-    @_comparator
     def __gt__(self, other):
-        return self.tolist() > other.tolist()
+        return compare_pvector(self, other, operator.gt)
 
-    @_comparator
     def __lt__(self, other):
-        return self.tolist() < other.tolist()
+        return compare_pvector(self, other, operator.lt)
 
-    @_comparator
     def __ge__(self, other):
-        return self.tolist() >= other.tolist()
+        return compare_pvector(self, other, operator.ge)
 
-    @_comparator
     def __le__(self, other):
-        return self.tolist() <= other.tolist()
+        return compare_pvector(self, other, operator.le)
 
     def __mul__(self, times):
         if times <= 0 or self is _EMPTY_PVECTOR:
@@ -1495,7 +1487,12 @@ _EMPTY_PBAG = _PBag(_EMPTY_PMAP)
 
 ######################################## Immutable object ##############################################
 
-def pclass(members='', name='PClass', verbose=False):
+def pclass(*args, **kwargs):
+    warnings.warn("'pclass' has been renamed to 'immutable' please use that name instead", DeprecationWarning)
+    return immutable(*args, **kwargs)
+
+
+def immutable(members='', name='Immutable', verbose=False):
     """
     Produces a class that either can be used standalone or as a base class for persistent classes.
 
@@ -1503,7 +1500,7 @@ def pclass(members='', name='PClass', verbose=False):
 
     Constructing a type and using it to instantiate objects:
 
-    >>> Point = pclass('x, y', name='Point')
+    >>> Point = immutable('x, y', name='Point')
     >>> p = Point(1, 2)
     >>> p2 = p.set(x=3)
     >>> p
@@ -1513,7 +1510,7 @@ def pclass(members='', name='PClass', verbose=False):
 
     Inheriting from a constructed type. In this case no type name needs to be supplied:
 
-    >>> class PositivePoint(pclass('x, y')):
+    >>> class PositivePoint(immutable('x, y')):
     ...     __slots__ = tuple()
     ...     def __new__(cls, x, y):
     ...         if x > 0 and y > 0:
@@ -1531,7 +1528,7 @@ def pclass(members='', name='PClass', verbose=False):
     cannot be updated. For example it could be used to implement an ID that should remain the same
     over time. A frozen member is denoted by a trailing underscore.
 
-    >>> Point = pclass('x, y, id_', name='Point')
+    >>> Point = immutable('x, y, id_', name='Point')
     >>> p = Point(1, 2, id_=17)
     >>> p.set(x=3)
     Point(x=3, y=2, id_=17)
@@ -1556,11 +1553,11 @@ def pclass(members='', name='PClass', verbose=False):
 
     quoted_members = ', '.join("'%s'" % m for m in members)
     template = """
-class {class_name}(namedtuple('PClassBase', [{quoted_members}], verbose={verbose})):
+class {class_name}(namedtuple('ImmutableBase', [{quoted_members}], verbose={verbose})):
     __slots__ = tuple()
 
     def __repr__(self):
-        return super({class_name}, self).__repr__().replace('PClassBase', self.__class__.__name__)
+        return super({class_name}, self).__repr__().replace('ImmutableBase', self.__class__.__name__)
 
     def set(self, **kwargs):
         if not kwargs:
@@ -1583,7 +1580,7 @@ class {class_name}(namedtuple('PClassBase', [{quoted_members}], verbose={verbose
         print(template)
 
     from collections import namedtuple
-    namespace = dict(namedtuple=namedtuple, __name__='pyrsistent_pclass')
+    namespace = dict(namedtuple=namedtuple, __name__='pyrsistent_immutable')
     try:
         six.exec_(template, namespace)
     except SyntaxError as e:
@@ -2402,12 +2399,13 @@ def field(type=_PRECORD_NO_TYPE, invariant=_PRECORD_NO_INVARIANT, initial=_PRECO
           mandatory=False, factory=_PRECORD_NO_FACTORY, serializer=_PRECORD_NO_SERIALIZER):
     """
     Field specification factory for :py:class:`PRecord`.
-    type, a type or iterable with types that are allowed for this field
-    invariant, a function specifying an invariant that must hold for the field
-    initial, initial value of field if not specified when instantiating the record
-    mandatorty, boolean specifying if the field is mandatory or not
-    factory, factory function called when field is set.
-    serializer, function that returns a serialized versino of the field
+
+    :param type: a type or iterable with types that are allowed for this field
+    :param invariant: a function specifying an invariant that must hold for the field
+    :param initial: value of field if not specified when instantiating the record
+    :param mandatorty: boolean specifying if the field is mandatory or not
+    :param factory: function called when field is set.
+    :param serializer: function that returns a serialized version of the field
     """
 
     types = set(type) if isinstance(type, Iterable) else set([type])
@@ -2505,7 +2503,7 @@ class PRecord(PMap, CheckedType):
 
     def __repr__(self):
         return "{0}({1})".format(self.__class__.__name__,
-                                 ', '.join('{0}={1}'.format(k, v) for k, v in self.items()))
+                                 ', '.join('{0}={1}'.format(k, repr(v)) for k, v in self.items()))
 
     @classmethod
     def create(cls, kwargs):
@@ -2536,6 +2534,25 @@ class PRecord(PMap, CheckedType):
 
         return dict((k, _serialize(k, v)) for k, v in self.items())
 
+
+class PRecordTypeError(TypeError):
+    """
+    Raised when trying to assign a value with a type that doesn't match the declared type.
+
+    Attributes:
+    source_class -- The class of the record
+    field -- Field name
+    expected_types  -- Types allowed for the field
+    actual_type -- The non matching type
+    """
+    def __init__(self, source_class, field, expected_types, actual_type, *args, **kwargs):
+        super(PRecordTypeError, self).__init__(*args, **kwargs)
+        self.source_class = source_class
+        self.field = field
+        self.expected_types = expected_types
+        self.actual_type = actual_type
+
+
 class _PRecordEvolver(PMap._Evolver):
     __slots__ = ('_destination_cls', '_invariant_error_codes', '_missing_fields')
 
@@ -2559,7 +2576,9 @@ class _PRecordEvolver(PMap._Evolver):
                 return self
 
             if field.type and not any(isinstance(value, t) for t in field.type):
-                raise TypeError("Invalid type for field {0}.{1}, was {2}".format(self._destination_cls.__name__, key, type(value)))
+                actual_type = type(value)
+                message = "Invalid type for field {0}.{1}, was {2}".format(self._destination_cls.__name__, key, actual_type.__name__)
+                raise PRecordTypeError(self._destination_cls, key, field.type, actual_type, message)
 
             is_ok, error_code = field.invariant(value)
             if not is_ok:
@@ -2688,13 +2707,16 @@ def _store_types(dct, bases, destination_name, source_name):
 
     dct[destination_name] = to_list(dct[source_name]) if source_name in dct else []
     dct[destination_name] += sum([to_list(b.__dict__[source_name]) for b in bases if source_name in b.__dict__], [])
+    dct[destination_name] = tuple(dct[destination_name])
     if not all(isinstance(t, type) for t in dct[destination_name]):
         raise TypeError('Type specifications must be types')
 
+
 def _store_invariants(dct, bases, destination_name, source_name):
         # Invariants are inherited
         dct[destination_name] = [dct[source_name]] if source_name in dct else []
         dct[destination_name] += [b.__dict__[source_name] for b in bases if source_name in b.__dict__]
+        dct[destination_name] = tuple(dct[destination_name])
         if not all(callable(invariant) for invariant in dct[destination_name]):
             raise TypeError('Invariants must be callable')
 
@@ -2716,9 +2738,51 @@ class _CheckedTypeMeta(type):
         return super(_CheckedTypeMeta, mcs).__new__(mcs, name, bases, dct)
 
 
-def _check_types(it, types):
-    if types and not all(any(isinstance(e, t) for t in types) for e in it):
-        raise TypeError
+class CheckedTypeError(TypeError):
+    def __init__(self, source_class, expected_types, actual_type, actual_value, *args, **kwargs):
+        super(CheckedTypeError, self).__init__(*args, **kwargs)
+        self.source_class = source_class
+        self.expected_types = expected_types
+        self.actual_type = actual_type
+        self.actual_value = actual_value
+
+
+class CheckedKeyTypeError(CheckedTypeError):
+    """
+    Raised when trying to set a value using a key with a type that doesn't match the declared type.
+
+    Attributes:
+    source_class -- The class of the collection
+    expected_types  -- Allowed types
+    actual_type -- The non matching type
+    actual_value -- Value of the variable with the non matching type
+    """
+    pass
+
+
+class CheckedValueTypeError(CheckedTypeError):
+    """
+    Raised when trying to set a value using a key with a type that doesn't match the declared type.
+
+    Attributes:
+    source_class -- The class of the collection
+    expected_types  -- Allowed types
+    actual_type -- The non matching type
+    actual_value -- Value of the variable with the non matching type
+    """
+    pass
+
+
+def _check_types(it, expected_types, source_class, exception_type=CheckedValueTypeError):
+    if expected_types:
+        for e in it:
+            if not any(isinstance(e, t) for t in expected_types):
+                actual_type = type(e)
+                msg = "Type {source_class} can only be used with {expected_types}, not {actual_type}".format(
+                    source_class=source_class.__name__,
+                    expected_types=tuple(et.__name__ for et in expected_types),
+                    actual_type=actual_type.__name__)
+                raise exception_type(source_class, expected_types, actual_type, e, msg)
 
 
 def _invariant_errors(elem, invariants):
@@ -2729,9 +2793,10 @@ def _invariant_errors_iterable(it, invariants):
     return sum([_invariant_errors(elem, invariants) for elem in it], [])
 
 
-def optional(typ):
-    """ Convenience function to specify that a value may be of type 'typ' or None """
-    return typ, type(None)
+def optional(*typs):
+    """ Convenience function to specify that a value may be of any of the types in type 'typs' or None """
+    return tuple(typs) + (type(None),)
+
 
 def _checked_type_create(cls, source_data):
         if isinstance(source_data, cls):
@@ -2750,6 +2815,17 @@ def _checked_type_create(cls, source_data):
 
 @six.add_metaclass(_CheckedTypeMeta)
 class CheckedPVector(_PVectorImpl, CheckedType):
+    """
+    A CheckedPVector is a PVector which allows specifying type and invariant checks.
+
+    >>> class Positives(CheckedPVector):
+    ...     __type__ = (long, int)
+    ...     __invariant__ = lambda n: (n >= 0, 'Negative')
+    ...
+    >>> Positives([1, 2, 3])
+    Positives([1, 2, 3])
+    """
+
     __slots__ = ()
 
     def __new__(cls, initial=()):
@@ -2786,7 +2862,7 @@ class CheckedPVector(_PVectorImpl, CheckedType):
             self._invariant_errors = []
 
         def _check(self, it):
-            _check_types(it, self._destination_class._checked_types)
+            _check_types(it, self._destination_class._checked_types, self._destination_class)
             error_data = _invariant_errors_iterable(it, self._destination_class._checked_invariants)
             self._invariant_errors.extend(error_data)
 
@@ -2822,8 +2898,20 @@ class CheckedPVector(_PVectorImpl, CheckedType):
     def evolver(self):
         return CheckedPVector.Evolver(self.__class__, self)
 
+
 @six.add_metaclass(_CheckedTypeMeta)
 class CheckedPSet(PSet, CheckedType):
+    """
+    A CheckedPVector is a PSet which allows specifying type and invariant checks.
+
+    >>> class Positives(CheckedPSet):
+    ...     __type__ = (long, int)
+    ...     __invariant__ = lambda n: (n >= 0, 'Negative')
+    ...
+    >>> Positives([1, 2, 3])
+    Positives([1, 2, 3])
+    """
+
     __slots__ = ()
 
     def __new__(cls, initial=()):
@@ -2840,7 +2928,7 @@ class CheckedPSet(PSet, CheckedType):
         return self.__class__.__name__ + super(CheckedPSet, self).__repr__()[4:]
 
     def __str__(self):
-       return self.__repr__()
+        return self.__repr__()
 
     def serialize(self, format=None):
         serializer = self.__serializer__
@@ -2852,7 +2940,6 @@ class CheckedPSet(PSet, CheckedType):
         # Pickling support
         return _restore_pickle, (self.__class__, list(self),)
 
-
     def evolver(self):
         return CheckedPSet.Evolver(self.__class__, self)
 
@@ -2865,7 +2952,7 @@ class CheckedPSet(PSet, CheckedType):
             self._invariant_errors = []
 
         def _check(self, it):
-            _check_types(it, self._destination_class._checked_types)
+            _check_types(it, self._destination_class._checked_types, self._destination_class)
             error_data = _invariant_errors_iterable(it, self._destination_class._checked_invariants)
             self._invariant_errors.extend(error_data)
 
@@ -2910,8 +2997,21 @@ class _CheckedMapTypeMeta(type):
 # Marker object
 _UNDEFINED_CHECKED_PMAP_SIZE = object()
 
+
 @six.add_metaclass(_CheckedMapTypeMeta)
 class CheckedPMap(PMap, CheckedType):
+    """
+    A CheckedPMap is a PMap which allows specifying type and invariant checks.
+
+    >>> class IntToFloatMap(CheckedPMap):
+    ...     __key_type__ = int
+    ...     __value_type__ = float
+    ...     __invariant__ = lambda k, v: (int(v) == k, 'Invalid mapping')
+    ...
+    >>> IntToFloatMap({1: 1.5, 2: 2.25})
+    IntToFloatMap({1: 1.5, 2: 2.25})
+    """
+
     __slots__ = ()
 
     def __new__(cls, initial={}, size=_UNDEFINED_CHECKED_PMAP_SIZE):
@@ -2968,20 +3068,18 @@ class CheckedPMap(PMap, CheckedType):
             self._invariant_errors = []
 
         def set(self, key, value):
-            _check_types([key], self._destination_class._checked_key_types)
-            _check_types([value], self._destination_class._checked_value_types)
+            _check_types([key], self._destination_class._checked_key_types, self._destination_class, CheckedKeyTypeError)
+            _check_types([value], self._destination_class._checked_value_types, self._destination_class)
             self._invariant_errors.extend(data for valid, data in (invariant(key, value)
                                                                    for invariant in self._destination_class._checked_invariants)
                                           if not valid)
 
             return super(CheckedPMap.Evolver, self).set(key, value)
 
-
         def persistent(self):
             if self._invariant_errors:
                 raise InvariantException(error_codes=self._invariant_errors)
 
-
             if self.is_dirty() or type(self._original_pmap) != self._destination_class:
                 return self._destination_class(self._buckets_evolver.persistent(), self._size)
 
